# Generated from ASLIntrinsicParser.g4 by ANTLR 4.13.1
from antlr4 import ParseTreeListener

if "." in __name__:
    from .ASLIntrinsicParser import ASLIntrinsicParser
else:
    from ASLIntrinsicParser import ASLIntrinsicParser


# This class defines a complete listener for a parse tree produced by ASLIntrinsicParser.
class ASLIntrinsicParserListener(ParseTreeListener):
    # Enter a parse tree produced by ASLIntrinsicParser#func_decl.
    def enterFunc_decl(self, ctx: ASLIntrinsicParser.Func_declContext):
        pass

    # Exit a parse tree produced by ASLIntrinsicParser#func_decl.
    def exitFunc_decl(self, ctx: ASLIntrinsicParser.Func_declContext):
        pass

    # Enter a parse tree produced by ASLIntrinsicParser#states_func_decl.
    def enterStates_func_decl(self, ctx: ASLIntrinsicParser.States_func_declContext):
        pass

    # Exit a parse tree produced by ASLIntrinsicParser#states_func_decl.
    def exitStates_func_decl(self, ctx: ASLIntrinsicParser.States_func_declContext):
        pass

    # Enter a parse tree produced by ASLIntrinsicParser#state_fun_name.
    def enterState_fun_name(self, ctx: ASLIntrinsicParser.State_fun_nameContext):
        pass

    # Exit a parse tree produced by ASLIntrinsicParser#state_fun_name.
    def exitState_fun_name(self, ctx: ASLIntrinsicParser.State_fun_nameContext):
        pass

    # Enter a parse tree produced by ASLIntrinsicParser#func_arg_list.
    def enterFunc_arg_list(self, ctx: ASLIntrinsicParser.Func_arg_listContext):
        pass

    # Exit a parse tree produced by ASLIntrinsicParser#func_arg_list.
    def exitFunc_arg_list(self, ctx: ASLIntrinsicParser.Func_arg_listContext):
        pass

    # Enter a parse tree produced by ASLIntrinsicParser#func_arg_string.
    def enterFunc_arg_string(self, ctx: ASLIntrinsicParser.Func_arg_stringContext):
        pass

    # Exit a parse tree produced by ASLIntrinsicParser#func_arg_string.
    def exitFunc_arg_string(self, ctx: ASLIntrinsicParser.Func_arg_stringContext):
        pass

    # Enter a parse tree produced by ASLIntrinsicParser#func_arg_int.
    def enterFunc_arg_int(self, ctx: ASLIntrinsicParser.Func_arg_intContext):
        pass

    # Exit a parse tree produced by ASLIntrinsicParser#func_arg_int.
    def exitFunc_arg_int(self, ctx: ASLIntrinsicParser.Func_arg_intContext):
        pass

    # Enter a parse tree produced by ASLIntrinsicParser#func_arg_float.
    def enterFunc_arg_float(self, ctx: ASLIntrinsicParser.Func_arg_floatContext):
        pass

    # Exit a parse tree produced by ASLIntrinsicParser#func_arg_float.
    def exitFunc_arg_float(self, ctx: ASLIntrinsicParser.Func_arg_floatContext):
        pass

    # Enter a parse tree produced by ASLIntrinsicParser#func_arg_bool.
    def enterFunc_arg_bool(self, ctx: ASLIntrinsicParser.Func_arg_boolContext):
        pass

    # Exit a parse tree produced by ASLIntrinsicParser#func_arg_bool.
    def exitFunc_arg_bool(self, ctx: ASLIntrinsicParser.Func_arg_boolContext):
        pass

    # Enter a parse tree produced by ASLIntrinsicParser#func_arg_context_path.
    def enterFunc_arg_context_path(
        self, ctx: ASLIntrinsicParser.Func_arg_context_pathContext
    ):
        pass

    # Exit a parse tree produced by ASLIntrinsicParser#func_arg_context_path.
    def exitFunc_arg_context_path(
        self, ctx: ASLIntrinsicParser.Func_arg_context_pathContext
    ):
        pass

    # Enter a parse tree produced by ASLIntrinsicParser#func_arg_json_path.
    def enterFunc_arg_json_path(
        self, ctx: ASLIntrinsicParser.Func_arg_json_pathContext
    ):
        pass

    # Exit a parse tree produced by ASLIntrinsicParser#func_arg_json_path.
    def exitFunc_arg_json_path(self, ctx: ASLIntrinsicParser.Func_arg_json_pathContext):
        pass

    # Enter a parse tree produced by ASLIntrinsicParser#func_arg_func_decl.
    def enterFunc_arg_func_decl(
        self, ctx: ASLIntrinsicParser.Func_arg_func_declContext
    ):
        pass

    # Exit a parse tree produced by ASLIntrinsicParser#func_arg_func_decl.
    def exitFunc_arg_func_decl(self, ctx: ASLIntrinsicParser.Func_arg_func_declContext):
        pass


del ASLIntrinsicParser
